@@ -2,7 +2,7 @@
  * Orthanc - A Lightweight, RESTful DICOM Store
  * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
  * Department, University Hospital of Liege, Belgium
- * Copyright (C) 2017-2019 Osimis S.A., Belgium
+ * Copyright (C) 2017-2020 Osimis S.A., Belgium
  *
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -38,6 +38,7 @@
 #include "../Core/Logging.h"
 #include "../Core/OrthancException.h"
 #include "../Core/SystemToolbox.h"
+#include "../Core/TemporaryFile.h"
 #include "../Core/Toolbox.h"
 
 #include "ServerIndex.h"
@@ -47,6 +48,7 @@ static const char* const DICOM_MODALITIES = "DicomModalities";
 static const char* const DICOM_MODALITIES_IN_DB = "DicomModalitiesInDatabase";
 static const char* const ORTHANC_PEERS = "OrthancPeers";
 static const char* const ORTHANC_PEERS_IN_DB = "OrthancPeersInDatabase";
+static const char* const TEMPORARY_DIRECTORY = "TemporaryDirectory";
 
 namespace Orthanc
 {
@@ -420,8 +422,8 @@ namespace Orthanc
   }
 
 
-  std::string OrthancConfiguration::GetStringParameter(const std::string& parameter,
-                                                       const std::string& defaultValue) const
+  bool OrthancConfiguration::LookupStringParameter(std::string& target,
+                                                   const std::string& parameter) const
   {
     if (json_.isMember(parameter))
     {
@@ -432,10 +434,26 @@ namespace Orthanc
       }
       else
       {
-        return json_[parameter].asString();
+        target = json_[parameter].asString();
+        return true;
       }
     }
     else
+    {
+      return false;
+    }
+  }
+
+
+  std::string OrthancConfiguration::GetStringParameter(const std::string& parameter,
+                                                       const std::string& defaultValue) const
+  {
+    std::string value;
+    if (LookupStringParameter(value, parameter))
+    {
+      return value;
+    }
+    else
     {
       return defaultValue;
     }
@@ -482,8 +500,8 @@ namespace Orthanc
   }
 
 
-  bool OrthancConfiguration::GetBooleanParameter(const std::string& parameter,
-                                                 bool defaultValue) const
+  bool OrthancConfiguration::LookupBooleanParameter(bool& target,
+                                                    const std::string& parameter) const
   {
     if (json_.isMember(parameter))
     {
@@ -495,10 +513,26 @@ namespace Orthanc
       }
       else
       {
-        return json_[parameter].asBool();
+        target = json_[parameter].asBool();
+        return true;
       }
     }
     else
+    {
+      return false;
+    }
+  }
+
+
+  bool OrthancConfiguration::GetBooleanParameter(const std::string& parameter,
+                                                 bool defaultValue) const
+  {
+    bool value;
+    if (LookupBooleanParameter(value, parameter))
+    {
+      return value;
+    }
+    else
     {
       return defaultValue;
     }
@@ -609,13 +643,13 @@ namespace Orthanc
   }
 
 
-  void OrthancConfiguration::SetupRegisteredUsers(HttpServer& httpServer) const
+  bool OrthancConfiguration::SetupRegisteredUsers(HttpServer& httpServer) const
   {
     httpServer.ClearUsers();
 
     if (!json_.isMember("RegisteredUsers"))
     {
-      return;
+      return false;
     }
 
     const Json::Value& users = json_["RegisteredUsers"];
@@ -624,13 +658,17 @@ namespace Orthanc
       throw OrthancException(ErrorCode_BadFileFormat, "Badly formatted list of users");
     }
 
+    bool hasUser = false;
     Json::Value::Members usernames = users.getMemberNames();
     for (size_t i = 0; i < usernames.size(); i++)
     {
       const std::string& username = usernames[i];
       std::string password = users[username].asString();
       httpServer.RegisterUser(username.c_str(), password.c_str());
+      hasUser = true;
     }
+
+    return hasUser;
   }
     
 
@@ -757,6 +795,8 @@ namespace Orthanc
   void OrthancConfiguration::UpdateModality(const std::string& symbolicName,
                                             const RemoteModalityParameters& modality)
   {
+    CheckAlphanumeric(symbolicName);
+    
     modalities_[symbolicName] = modality;
     SaveModalities();
   }
@@ -772,6 +812,8 @@ namespace Orthanc
   void OrthancConfiguration::UpdatePeer(const std::string& symbolicName,
                                         const WebServiceParameters& peer)
   {
+    CheckAlphanumeric(symbolicName);
+    
     peer.CheckClientCertificate();
 
     peers_[symbolicName] = peer;
@@ -826,4 +868,24 @@ namespace Orthanc
   {
     serverIndex_ = NULL;
   }
+
+  
+  TemporaryFile* OrthancConfiguration::CreateTemporaryFile() const
+  {
+    if (json_.isMember(TEMPORARY_DIRECTORY))
+    {
+      return new TemporaryFile(InterpretStringParameterAsPath(GetStringParameter(TEMPORARY_DIRECTORY, ".")), "");
+    }
+    else
+    {
+      return new TemporaryFile;
+    }
+  }
+
+
+  std::string OrthancConfiguration::GetDefaultPrivateCreator() const
+  {
+    // New configuration option in Orthanc 1.6.0
+    return GetStringParameter("DefaultPrivateCreator", "");
+  }
 }
