@@ -2,7 +2,7 @@
  * Orthanc - A Lightweight, RESTful DICOM Store
  * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
  * Department, University Hospital of Liege, Belgium
- * Copyright (C) 2017-2019 Osimis S.A., Belgium
+ * Copyright (C) 2017-2020 Osimis S.A., Belgium
  *
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -38,6 +38,7 @@
 #include "../Core/DicomParsing/FromDcmtkBridge.h"
 #include "../Core/Logging.h"
 #include "../Core/Lua/LuaFunctionCall.h"
+#include "../Core/MetricsRegistry.h"
 #include "OrthancConfiguration.h"
 #include "Search/DatabaseLookup.h"
 #include "ServerContext.h"
@@ -272,7 +273,7 @@ namespace Orthanc
       throw OrthancException(ErrorCode_UnknownResource);  // The resource was deleted in between
     }
 
-    std::auto_ptr<DicomMap> result(new DicomMap);
+    std::unique_ptr<DicomMap> result(new DicomMap);
 
     switch (level)
     {
@@ -370,7 +371,7 @@ namespace Orthanc
     }
     else
     {
-      ParsedDicomFile dicom(result);
+      ParsedDicomFile dicom(result, GetDefaultDicomEncoding(), true /* be permissive, cf. issue #136 */);
 
       for (std::list<DicomTag>::const_iterator tag = sequencesToReturn.begin();
            tag != sequencesToReturn.end(); ++tag)
@@ -393,7 +394,8 @@ namespace Orthanc
             content.append(item);
           }
 
-          dicom.Replace(*tag, content, false, DicomReplaceMode_InsertIfAbsent);
+          dicom.Replace(*tag, content, false, DicomReplaceMode_InsertIfAbsent,
+                        "" /* no private creator */);
         }
       }
 
@@ -535,7 +537,7 @@ namespace Orthanc
                        const DicomMap& mainDicomTags,
                        const Json::Value* dicomAsJson) 
     {
-      std::auto_ptr<DicomMap> counters(ComputeCounters(context_, instanceId, level_, query_));
+      std::unique_ptr<DicomMap> counters(ComputeCounters(context_, instanceId, level_, query_));
 
       AddAnswer(answers_, mainDicomTags, dicomAsJson,
                 queryAsArray_, sequencesToReturn_, counters.get());
@@ -551,6 +553,8 @@ namespace Orthanc
                                          const std::string& calledAet,
                                          ModalityManufacturer manufacturer)
   {
+    MetricsRegistry::Timer timer(context_.GetMetricsRegistry(), "orthanc_find_scp_duration_ms");
+
     /**
      * Possibly apply the user-supplied Lua filter.
      **/
@@ -639,7 +643,7 @@ namespace Orthanc
       std::string value = element.GetValue().GetContent();
       if (value.size() == 0)
       {
-        // An empty string corresponds to a "*" wildcard constraint, so we ignore it
+        // An empty string corresponds to an universal constraint, so we ignore it
         continue;
       }
 
